/* stream.c   by Michael Thorpe   2010-08-11 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* Just for isatty(3) */
#include "lisp.h"

static obj *eoferror,*ioerror;

static obj *close_file(obj *stream) {
	if(stream->type != &stream_objtype)
		return(throwtypeerror(stream));
	if(!stream->value.stream.closable)
		return(throwaccesserror(stream));
	if(fclose(stream->value.stream.f))
		return(throw(INCREF(ioerror),stream));
	stream->value.stream.readable=0;
	stream->value.stream.writable=0;
	stream->value.stream.closable=0;
	return(stream);
}

static obj *isatty_func(obj *stream) {
	obj *o;

	if(&stream_objtype != stream->type)
		return(throwtypeerror(stream));
	if(!stream->value.stream.readable && !stream->value.stream.writable)
		return(throwaccesserror(stream));
	if(isatty(fileno(stream->value.stream.f)))
		o=&trueobj;
	else
		o=&nilobj;
	DECREF(stream);
	return(INCREF(o));
}

static obj *open_file(obj *name,obj *mode) {
	obj *o;
	int i;
	char n[PATH_MAX+1];
	char m[3];

	if(name->type != &string_objtype) {
		DECREF(mode);
		return(throwtypeerror(name));
	}
	if(mode->type != &string_objtype || mode->value.s.len>2) {
		DECREF(name);
		return(throwtypeerror(mode));
	}
	if(name->value.s.len>PATH_MAX) {
		DECREF(mode);
		return(throwrangecheck(name));
	}
	memcpy(m,mode->value.s.data,mode->value.s.len);
	if('a' != m[0] && 'r' != m[0] && 'w' != m[0]) {
		DECREF(name);
		return(throwtypeerror(mode));
	}
	if(mode->value.s.len>2) {
		DECREF(name);
		return(throwargumenterror(mode));
	}
	if(2==mode->value.s.len && '+' != m[1]) {
		DECREF(name);
		return(throwtypeerror(mode));
	}
	m[mode->value.s.len]='\0';
	for(i=0;i<name->value.s.len;i++) {
		if(!name->value.s.data[i]) {
			DECREF(mode);
			return(throwargumenterror(name));
		}
		n[i]=name->value.s.data[i];
	}
	n[i]='\0';
	DECREF(mode);
	o=new_object(&stream_objtype);
	if(!o) {
		DECREF(name);
		return(0);
	}
	o->value.stream.readable=0;
	o->value.stream.writable=0;
	o->value.stream.closable=0;
	o->value.stream.f=fopen(n,m);
	if(!o->value.stream.f) {
		DECREF(o);
		return(throwaccesserror(name));
	}
	DECREF(name);
	if('r'==m[0] || '+'==m[1])
		o->value.stream.readable=1;
	if('r' != m[0] || '+'==m[1])
		o->value.stream.writable=1;
	o->value.stream.closable=1;
	return(o);
}

static obj *sgetc(obj *stream) {
	int c;
	obj *o;

	if(stream->type != &stream_objtype)
		return(throwtypeerror(stream));
	if(!stream->value.stream.readable)
		return(throwaccesserror(stream));
	o=new_object(&integer_objtype);
	if(o) {
		c=fgetc(stream->value.stream.f);
		if(EOF==c) {
			if(feof(stream->value.stream.f)) {
				DECREF(o);
				return(throw(INCREF(eoferror),stream));
			} else {
				DECREF(o);
				return(throw(INCREF(ioerror),stream));
			}
		}
		o->value.i=c;
	}
	DECREF(stream);
	return(o);
}

static obj *speekc(obj *stream) {
	obj *o;
	int c;

	if(stream->type != &stream_objtype)
		return(throwtypeerror(stream));
	if(!stream->value.stream.readable)
		return(throwaccesserror(stream));
	o=new_object(&integer_objtype);
	if(o) {
		c=fgetc(stream->value.stream.f);
		if(EOF==c) {
			if(feof(stream->value.stream.f)) {
				DECREF(o);
				return(throw(INCREF(eoferror),stream));
			} else {
				DECREF(o);
				return(throw(INCREF(ioerror),stream));
			}
		}
		if(EOF==ungetc(c,stream->value.stream.f)) {
			DECREF(o);
			return(throw(INCREF(ioerror),stream));
		}
		o->value.i=c;
	}
	DECREF(stream);
	return(o);
}

static obj *sputc(obj *stream,obj *c) {
	int i;

	if(stream->type != &stream_objtype) {
		DECREF(c);
		return(throwtypeerror(stream));
	}
	if(c->type != &integer_objtype) {
		DECREF(stream);
		return(throwtypeerror(c));
	}
	if(!stream->value.stream.writable) {
		DECREF(c);
		return(throwaccesserror(stream));
	}
	if(c->value.i<0 || c->value.i>UCHAR_MAX) {
		DECREF(stream);
		return(throwrangecheck(c));
	}
	i=fputc((int)c->value.i,stream->value.stream.f);
	if(EOF==i) {
		DECREF(c);
		return(throw(INCREF(ioerror),stream));
	}
	DECREF(stream);
	return(c);
}

static void stream_dropobj(obj *in) {
	if(in->value.stream.closable)
		fclose(in->value.stream.f);
}

static obj *stream_tostring(obj *in) {
	obj *out;
	char buf[6+1+2+16+1];
	int i;

	i=snprintf(buf,sizeof(buf),"stream-%p",in->value.stream.f);
	if(i==-1 || i>=sizeof(buf))
		return(throwrangecheck(in));
	DECREF(in);
	out=new_string(buf,i);
	return(out);
}

static obj *write_string(obj *stream,obj *string) {
	unsigned long i,j;

	if(stream->type != &stream_objtype) {
		DECREF(string);
		return(throwtypeerror(stream));
	}
	if(string->type != &string_objtype) {
		DECREF(stream);
		return(throwtypeerror(string));
	}
	if(!stream->value.stream.writable) {
		DECREF(string);
		return(throwaccesserror(stream));
	}
	for(i=0;i<string->value.s.len;i+=j) {
		j=fwrite(string->value.s.data,1,string->value.s.len,stream->value.stream.f);
		if(-1==j || 0==j) {
			DECREF(string);
			return(throw(INCREF(ioerror),stream));
		}
	}
	DECREF(stream);
	return(string);
}

static int stream_init() {
	obj *in,*out,*err;

	if(!(eoferror=new_word("end-of-file",11))
	|| !(ioerror=new_word("io-error",8))
	|| !(in=new_object(&stream_objtype))
	|| !(out=new_object(&stream_objtype))
	|| !(err=new_object(&stream_objtype))
	|| store_builtin_ns1("close-file",close_file)
	|| store_builtin_ns1("isatty",isatty_func)
	|| store_builtin_ns2("open-file",open_file)
	|| store_builtin_ns1("read-char",sgetc)
	|| store_builtin_ns1("peek-char",speekc)
	|| store_builtin_ns2("write-char",sputc)
	|| store_builtin_ns2("write-string",write_string))
		return(1);
	in->value.stream.readable=1;
	in->value.stream.writable=0;
	in->value.stream.closable=0;
	in->value.stream.f=stdin;
	out->value.stream.readable=0;
	out->value.stream.writable=1;
	out->value.stream.closable=0;
	out->value.stream.f=stdout;
	err->value.stream.readable=0;
	err->value.stream.writable=1;
	err->value.stream.closable=0;
	err->value.stream.f=stderr;
	if(store_object("#standard-in",in)
	|| store_object("#standard-out",out)
	|| store_object("#standard-error",err))
		return(1);
	return(0);
}

static void stream_exit() {
	DECREF(eoferror); eoferror=0;
	DECREF(ioerror); ioerror=0;
}

objtype stream_objtype={
	name: "stream",
	tostring: stream_tostring,
	predropobj: stream_dropobj,
	init: stream_init,
	exit: stream_exit,
};

